# -*- coding: utf-8 -*-
# Diese __init__ ist dazu geschrieben worden, damit Module oder Pakete die
# nach dem disutils-layout gestaltet sind, in Umgebungen verwendet werden
# können, wo ein benutzerdefinierter Pfad auf zu verwendende Module oder
# Pakete zeigt. Das Problem dabei mit dem disutils-layout ist, dass Python
# solche Pakete nicht automatisch sauber einbindet. Für das Beispielpaket
# userdefinedpackage würde der Aufbau so aussehen:
#
#     ./userdefinedpackage/setup.py
#     ./userdefinedpackage/userdefinedpackage/__init__.py
#     ./userdefinedpackage/userdefinedpackage/custommodule.py
#
# Def Befehl:
#
#     import userdefinedpackage
#
# würde den Ordner ./userdefinedpackage/ im benutzerdefinierten Pfad als
# Namespace einbinden. Das Paket oder Module befindet sich jedoch im Pfad
# ./userdefinedpackage/userdefinedpackage.
#
# Dies wird hier dadurch gelöst, dass dieses Paketverzeichnis in die aktuelle
# Laufzeitumgebung vor den benutzerdefinierten Paketpfad eingebunden wird.
# Dadurch wird in der import Abfrage als erstes dieses Paket gefunden und
# importiert.
import os
import sys
from importlib import reload, import_module

_currentpackagepath = os.path.normpath(os.path.split(__file__)[0])
_currentpackagename = os.path.split(_currentpackagepath)[1]
if _currentpackagepath not in sys.path:
    print("Adding {} to current runtime paths".format(_currentpackagepath))
    sys.path.insert(0, _currentpackagepath)

    del sys.modules[_currentpackagename]
    import_module(_currentpackagename)



# -*- coding: utf-8 -*-
"""
Basic Usage of **JsonModel**
============================

JsonModel is designed to pre-define a dictionary with type testing and conversion. Its
secondary task is to eliminate occuring magic constants of dictionary keys.

.. testsetup::

   from jsonmetamodel import JsonModel, JsonInteger, JsonString

class definition
----------------

A JsonModel is defined by declaring attributes with *IDataType*, which defines

.. testcode::

    class Person(JsonModel):
        name = JsonString(default="Not Sure")
        age = JsonInteger(default=538)

    print(Person())

.. testoutput::

    MappingTreeItem({'name': 'Not Sure', 'age': 538})


access to dictionary key values
-------------------------------

The key values can be obtained by *attribute_name* of the declared *IDataType*
attributes of a JsonModel.

.. testcode::

    richard_watterson = Person({'age': 538, 'name': 'Richard Watterson'})
    richard_watterson[Person.age.attribute_name] = "38"
    print(richard_watterson[Person.age.attribute_name])

.. testoutput::

    38

The JsonModel-class
===================

.. autoclass:: JsonModel
   :members:

"""

import logging
import copy
from jsonmetamodel.fieldtypes import ADataType
from augmentedtree import (
    augment_datastructure,
    MappingTreeItem,
)
from typing import Mapping, Optional, Dict
from dicthandling import keep_keys


LOGGER = logging.getLogger("jsonmetamodel")


# The custom dictionary
class _member_table(dict):
    """
    Class member table for the Model class.

    Attributes
    ----------
    :Fields: Contains IDataType Objects as column data type definitions.
    :Defaults: Contains a dictionary with the default values.
    :KeyPairs: Contains a dictionary with the Attribute-Key and its column key
    :ColumnNames: List of column names.
    """

    def __init__(self):
        self.field_types: dict = {}
        self.defaults: dict = {}
        self.attribute_names: dict = {}
        self.primarykeys: dict = {}

    def __setitem__(self, attribute_name, attribute_value):
        # if the key is not already defined, add to the
        # list of keys.
        if isinstance(attribute_value, ADataType):
            # if no different columnname was defined the attribute name is
            # used as the dictionary key.
            attribute_value.set_attribute_name(attribute_name)
            attributes_destination_name = attribute_value.primekey
            self.field_types[attributes_destination_name] = attribute_value
            self.attribute_names[attribute_name] = attributes_destination_name
            self.primarykeys[attributes_destination_name] = attribute_name
            # Call superclass to add Additional attribute
            # dict.__setitem__(self, attribute_name, value_or_idatatype.columnname)
            self.defaults[attributes_destination_name] = attribute_value.default
        dict.__setitem__(self, attribute_name, attribute_value)


class JsonModelClass(type):
    # The prepare function
    @classmethod
    def __prepare__(mcs, name, bases):  # No keywords in this case
        finalTable = _member_table()
        for base in bases:
            finalTable.field_types.update(base.field_types)
            finalTable.defaults.update(base.defaults)
            finalTable.attribute_names.update(base.attribute_names)
            finalTable.primarykeys.update(base.primarykeys)
        return finalTable

    # The metaclass invocation
    def __new__(mcs, name, bases, classdict):
        # Note that we replace the classdict with a regular
        # dict before passing it to the superclass, so that we
        # don't continue to record member names after the class
        # has been created.
        _result = type.__new__(mcs, name, bases, dict(classdict))
        _result.defaults = classdict.defaults
        _result.field_types = classdict.field_types
        _result.attribute_names = classdict.attribute_names
        _result.primarykeys = classdict.primarykeys
        return _result


class JsonModel(metaclass=JsonModelClass):
    def __new__(
        cls, data: Mapping = None, use_model_keyorder: bool = False, **kwargs
    ) -> MappingTreeItem:
        """
        The **JsonModel**-class is an interface like definition of dictionaries
        with type checking and conversion capabilities.

        Examples:

            >>> from jsonmetamodel import JsonModel, JsonInteger, JsonString
            >>> class Person(JsonModel):
            ...     name = JsonString(default="Not Sure")
            ...     age = JsonInteger(default=538)
            ...
            >>> Person()
            MappingTreeItem({'name': 'Not Sure', 'age': 538})

            >>> class ABunchOfItems(JsonModel):
            ...     any = JsonString()
            ...     thing = JsonString()
            ...     here = JsonString()
            ...     and_a = JsonModel()
            ...     lot = JsonModel()
            ...
            >>> ABunchOfItems(any='value and', lot='more')
            MappingTreeItem({'any': 'value and', 'thing': '', 'here': '', 'lot': 'more'})

        Args:
            data (Mapping):
                This given data as a Mapping will update the predefined fields
                of this class.

            use_model_keyorder (bool):
                States if the key order of this class will be forced upon the
                result. If not, the order of the incoming data is used.

            **kwargs: iterable if data is given.

        Returns:
            MappingTreeItem (Mapping)
        """
        return cls._get_treeitem(data, use_model_keyorder, **kwargs)

    @classmethod
    def _get_keypairs_of_different_associations(cls):
        result = {}
        for attributen_name, primarykey in cls.attribute_names.items():
            if attributen_name != primarykey:
                result[attributen_name] = primarykey
        return result

    @classmethod
    def _override_data_with_keyword_arguments(
        cls, data: Optional[Mapping], overriding_kwargs: dict
    ) -> dict:
        """
        .. doctest::
           :hide:

            >>> from jsonmetamodel import JsonModel, JsonInteger, JsonString
            >>> class Person(JsonModel):
            ...     name = JsonString(default="Not Sure")
            ...     age = JsonInteger(default=538)
            ...
            >>> sample_data = {"something": "not_in_person", "age": 1}
            >>> fake_kwargs = {"age": 4}
            >>> Person._override_data_with_keyword_arguments(sample_data, fake_kwargs)
            {'name': 'Not Sure', 'age': 4, 'something': 'not_in_person'}

        Args:
            data(Optional[Mapping]):
                Mapping which is going to be overriden by additional keyword
                arguments.

            overriding_kwargs(Dict):
                Given keyword arguments, which override the given data.

        Returns:
            dict
        """
        required_defaults = cls.defaults.copy()
        if data is not None:
            required_defaults.update(data)
        required_defaults.update(overriding_kwargs)
        return required_defaults

    @classmethod
    def _retrieve_default(cls) -> MappingTreeItem:
        """
        .. doctest::
           :hide:

            >>> from jsonmetamodel import JsonModel, JsonInteger, JsonString
            >>> class Person(JsonModel):
            ...     name = JsonString(default="Not Sure")
            ...     age = JsonInteger(default=538)
            ...
            >>> Person._retrieve_default()
            MappingTreeItem({'name': 'Not Sure', 'age': 538})

        Returns:
            MappingTreeItem
        """
        default_data = cls.defaults.copy()
        treeitem: MappingTreeItem = augment_datastructure(nested_data=default_data, )
        treeitem.set_field_types(cls.field_types.copy())
        return treeitem

    @classmethod
    def _get_treeitem(
        cls, data: Mapping, use_model_keyorder: bool, **kwargs
    ) -> MappingTreeItem:
        do_retrieve_complete_default = (data is None) and (not kwargs)
        if do_retrieve_complete_default:
            return cls._retrieve_default()

        overridden_data = cls._override_data_with_keyword_arguments(data, kwargs)
        requested_names = set(list(overridden_data))
        needed_names = set(cls.primarykeys)
        fieldnames_to_add = needed_names.difference(requested_names)
        fieldnames_to_convert = requested_names.intersection(needed_names)

        for fieldname in fieldnames_to_convert:
            try:
                value_to_convert = overridden_data[fieldname]
                converter = cls.field_types[fieldname]
                overridden_data[fieldname] = converter(value_to_convert)
            except ValueError:
                errormsg = (
                    "Given value for field `{}` could not be converted into"
                    " destined type `{}`.".format(fieldname, cls.field_types[fieldname])
                )
                raise ValueError(errormsg)

        if use_model_keyorder:
            resulting_data = cls.defaults.copy()
            resulting_data.update(overridden_data)
        else:
            for fieldname in fieldnames_to_add:
                overridden_data[fieldname] = cls.defaults[fieldname]
            resulting_data = overridden_data

        different_keypairs = cls._get_keypairs_of_different_associations()
        treeitem_with_modelattributes = augment_datastructure(
            nested_data=resulting_data,
        )
        treeitem_with_modelattributes.set_field_types(cls.field_types.copy())
        treeitem_with_modelattributes.set_keypairs_of_related_values(different_keypairs)
        return treeitem_with_modelattributes

    @classmethod
    def extract(cls, data: Mapping, checkgroup_name: str = None) -> MappingTreeItem:
        """
        Returns an *AttributedMappingTreeItem* of this class with the
        values of *data* for keys which fit to this class.

        Args:
            data (dict):
                Data which shall be used.

            checkgroup_name (str; optional):
                default = None; Group which should be extracted.

        Returns:
            AttributedMappingTreeItem:
                Intersection between data and this JsonModel's attributes.
        """
        if not isinstance(data, dict):
            raise ValueError("Given data has to be a dictionary.")
        field_names_to_keep = cls.get_checkgroup_membernames(checkgroup_name)
        result = keep_keys(data, field_names_to_keep)
        return cls(result)

    @classmethod
    def intersect(cls, incoming_data: dict):
        """
        Intersection of fields within the `incoming_data` and this
        JsonModels definition.

        Args:
            incoming_data (dict):
                Data which fields intersecting with this JsonModel
                should remain.
        Returns:
            MappingTreeItem:
                Returning fields, which are fields of this JsonModel,
                but does not fill in missing fields into data.

        Examples:

            >>> from jsonmetamodel import JsonModel, JsonString
            >>> class Fruit(JsonModel):
            ...     color = JsonString()
            ...     type = JsonString()
            ...
            >>> apple = {'type': 'apple', 'color': 'red', 'variety': 'Golden Delicious'}
            >>> Fruit.intersect(apple)
            MappingTreeItem({'type': 'apple', 'color': 'red'})

        """
        if not isinstance(incoming_data, Mapping):
            raise ValueError("Given data has to be a Mapping.")
        remaining_data = copy.deepcopy(incoming_data)
        existing_field_names = list(incoming_data)
        for fieldname in existing_field_names:
            if fieldname not in cls.primarykeys:
                remaining_data.pop(fieldname)

        different_keypairs = cls._get_keypairs_of_different_associations()
        treeitem_with_modelattributes = augment_datastructure(
            nested_data=remaining_data,
        )
        treeitem_with_modelattributes.set_field_types(cls.field_types.copy())
        treeitem_with_modelattributes.set_keypairs_of_related_values(different_keypairs)
        return treeitem_with_modelattributes

    @classmethod
    def get_blank(cls) -> dict:
        """
        Returns a dictionary with the default values as a copy.

        Returns:
            dict:
                Dictionary with default values of this JsonModel

        Examples:

            >>> from jsonmetamodel import JsonModel, JsonInteger, JsonString
            >>> class Person(JsonModel):
            ...     name = JsonString(default="Not Sure")
            ...     age = JsonInteger(default=538)
            ...
            >>> Person.get_blank()
            {'name': 'Not Sure', 'age': 538}

        """
        return cls.defaults.copy()

    @classmethod
    def check(cls, data: Mapping, checkgroup_name: str = None) -> bool:
        """
        Checks if given data fits to the defined fields of this
        JsonModel and/or specific checkgroup.

        Args:
            data (Mapping):
                Data which should be checked.

            checkgroup_name (str; optional):
                default = None; blockname of 'checkgroup' which should be
                used.

        Returns:
            boolean
        """
        if not isinstance(data, Mapping):
            return False

        mandatory_field_names = cls.get_checkgroup_membernames(checkgroup_name)

        for typed_attribute_name in mandatory_field_names:
            if typed_attribute_name not in data:
                return False
        return True

    @classmethod
    def get_checkgroup_membernames(cls, checkgroup_name):
        if checkgroup_name is None:
            return cls.attribute_names
        checkgroup_member_names = []
        for typed_attribute_name, type_field in cls.field_types.items():
            if checkgroup_name not in type_field.checkgroups:
                continue
            checkgroup_member_names.append(typed_attribute_name)
        if not checkgroup_member_names:
            raise KeyError(
                "Checkgroup `{}` doesn't exist within "
                "{}".format(checkgroup_name, cls.__name__)
            )
        return checkgroup_member_names


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
.. autoclass:: jsonmetamodel.IDataType

.. autoclass:: jsonmetamodel.SqlDouble

.. autoclass:: jsonmetamodel.JsonInteger

.. autoclass:: jsonmetamodel.SqlInteger

.. autoclass:: jsonmetamodel.SqlBoolean

.. autoclass:: jsonmetamodel.JsonString

.. autoclass:: jsonmetamodel.SqlChar
"""
import logging
from typing import Any, List, Union, Callable, Optional, Tuple
import warnings
from abc import ABC, abstractmethod

logger = logging.getLogger("jsonmetamodel")


SQLCHAR_MAXIMUM_STRING_LENGTH = 255

SQL_INT_MINIMUM = -2147483648
SQL_INT_MAXIMUM = 2147483647


ValueIsValid = bool
ValidationErrorMessage = str
ValidationResult = Tuple[ValueIsValid, ValidationErrorMessage]


class ADataType(ABC):
    """"""

    def __init__(
        self,
        default: Optional[Any] = None,
        checkgroups: Optional[List[str]] = None,
        primekey: Optional[str] = None,
        allows_none=True,
        **additional_arguments
    ):
        """
        Basic field to define data types within a *jsonmetamodel.JsonModel*.

        Args:
            default:
            checkgroups(Optional[str, List[str]]):
                Declares the group name(s) this field is assigned to. Default
                is no assignment represented as an empty list.

            primekey(Optional[str]):
                The key name the field is assigned within the resulting
                dictionary. It can differ from the attributes name, which
                is the default value, if *primekey* is not assigned.

            allows_none(bool):
                States if 'None' values are allowed for this field. Default
                behaviour allows 'None' values. This is important, since if
                'None' values are forbidden, the value needs to be converted
                into the targeted data type.

            **additional_arguments
        """
        self._attribute_name = None
        self._primekey = primekey
        if checkgroups is None:
            self._checkgroups = ()
        elif isinstance(checkgroups, str):
            self._checkgroups = (checkgroups,)
        elif isinstance(checkgroups, (list, tuple)):
            self._checkgroups = checkgroups
        else:
            raise TypeError("'checkgroups' must be a list")
        self._default_value = None
        self._default_value = self(default)
        self.allows_none = allows_none
        self._additional_arguments = dict(additional_arguments)

    @property
    def primekey(self):
        if self._primekey is None:
            return self._attribute_name
        return self._primekey

    @property
    def attribute_name(self):
        return self._attribute_name

    def set_attribute_name(self, attribute_name: str):
        """
        Defines in which attribute of an `JsonModel` this object is
        declared.

        Args:
            attribute_name (str):
                blockname of attribute in a `JsonModel`

        Returns:
            bool:
                If name was set.
        """
        if self._attribute_name is None:
            self._attribute_name = attribute_name

    def grab(self, key, failedreturn=None):
        try:
            result = self._additional_arguments[key]
            return result
        except KeyError:
            return failedreturn

    def put(self, key, value):
        """
        Puts a key-value pair into the additional arguments of this field.

        Args:
            key(Hashable):

            value(Any):

        Returns:

        """
        self._additional_arguments[key] = value

    @abstractmethod
    def validate(self, value_to_validate) -> ValidationResult:
        """
        Validates the value, returning a *ValidationResult*.

        Args:
            value_to_validate(Any):
                The type which is validated within the scope of this field.

        Returns:
            ValidationResult(Tuple[bool, str]):
                Tuple of *ValueIsValid* stating if the value is within the
                bounds of this field and *ValidationErrorMessage* explaining
                the reason if its not the case.
        """
        pass

    @property
    def default(self):
        return self._default_value

    @property
    def checkgroups(self):
        """
        A list of groups this attribute is assigned to.

        Returns:
            list of str:
                Group names this attribute is assigned to.
        """
        return self._checkgroups

    @abstractmethod
    def __call__(self, incoming_data):
        """
        Converts `incoming_data` to the destined data type of this
        class.

        Args:
            incoming_data(Any):
                Data, which should be converted.

        Raises:
            ValueError:
                If `incoming_data` could not be converted into the
                destined data type of this class.

        Returns:
            Destined data type.
        """
        pass

    def to_targettype(self, value):
        """
        Does nothing, but passes the value through. This functionality is
        proposed for objects, which are no built in data types.
        """
        warnings.warn("Deprecated method; use class-call instead.", DeprecationWarning)
        return self.__call__(value)

    def encode(self, value):
        """
        Does nothing, but passes the value through. This functionality is
        proposed for objects, which are no built in data types.
        """
        warnings.warn(
            "Deprecated method; use class-call instead.", DeprecationWarning
        )
        return value


class IDataType(ADataType):
    """
    This class is deprecated and will be replaced by ADataType.

    """

    def __init__(
        self,
        default: Optional[Any] = None,
        checkgroups: Optional[List[str]] = None,
        primekey: Optional[str] = None,
        allows_none=True,
        **additional_arguments
    ):
        """
        Basic field to define data types within a *jsonmetamodel.JsonModel*.

        Args:
            default(Optional[Any]):
                The default value for this field.

            checkgroups(Optional[str, List[str]]):
                Declares the group name(s) this field is assigned to. Default
                is no assignment represented as an empty list.

            primekey(Optional[str]):
                The key name the field is assigned within the resulting
                dictionary. It can differ from the attributes name, which
                is the default value, if *primekey* is not assigned.

            allows_none(bool):
                States if 'None' values are allowed for this field. Default
                behaviour allows 'None' values. This is important, since if
                'None' values are forbidden, the value needs to be converted
                into the targeted data type.

            **additional_arguments
        """
        warnings.warn("IDataType is replaced by ADataType", DeprecationWarning)
        warnings.warn("IDataType is replaced by ADataType", FutureWarning)
        super().__init__(
            default=default,
            checkgroups=checkgroups,
            primekey=primekey,
            allows_none=allows_none,
        )


def returns_none_if_allowed(
    idatatype_call_function: Callable[[ADataType, Any], Any]
) -> Callable[[ADataType, Any], Any]:
    """
    Wrapper for functions returning 'None' values directly if the
    IDataType-Instance of the given *idatatype_call_function* allows them.
    If not a ValueError is raised.

    Raises:
        ValueError:
            Raised if 'None' values are not allowed, but received.

    Args:
        idatatype_call_function(Callable[[IDataType, Any], Any]):
            Conversion function of the parent IDataType-Instance.

    Returns:
        Callable[[IDataType, Any], Any]:
            check_if_none_is_allowed(
                this_idatatype_instance: IDataType,
                incoming_data_for_conversion
            )
    """

    def check_if_none_is_allowed(
        this_idatatype_instance: ADataType, incoming_data_for_conversion
    ):
        if incoming_data_for_conversion is None:
            if this_idatatype_instance.allows_none:
                return incoming_data_for_conversion
            raise ValueError(
                "'{}' does not allows 'None' values. You need to define an value "
                "fitting to {}.".format(
                    this_idatatype_instance.attribute_name,
                    this_idatatype_instance.__class__.__name__,
                )
            )
        return idatatype_call_function(
            this_idatatype_instance, incoming_data_for_conversion
        )

    return check_if_none_is_allowed


class JsonObject(ADataType):
    def __call__(self, incoming_data):
        return incoming_data

    def validate(self, value_to_validate) -> ValidationResult:
        raise NotImplementedError("validate is not implemented within this version.")


class AnyType(ADataType):
    """
    This should pass any data type or value through.

    Examples:

        >>> values_of_any_type = AnyType()
        >>> values_of_any_type(1)
        1
        >>> values_of_any_type("2")
        '2'
        >>> values_of_any_type(dict())
        {}
    """

    def __call__(self, incoming_data):
        return incoming_data

    def validate(self, value_to_validate) -> ValidationResult:
        return (True, "")


class SqlDouble(ADataType):
    """
    Examples:

        from jsonmetamodel import SqlDouble
        >>> a_floating_point_number = SqlDouble(default=0.0)
        >>> a_floating_point_number(1.2345)
        1.2345
        >>> a_floating_point_number("1.2345")
        1.2345
        >>> a_floating_point_number(1/100)
        0.01
    """

    def __init__(self, default: Optional[float] = 0.0, **additional_arguments):
        """

        Args:
            default(Optional[float]):
                The default value for this field.

            checkgroups(Optional[str, List[str]]):
                See ADataType

            primekey(Optional[str]):
                See ADataType

            allows_none(bool):
                See ADataType

            **additional_arguments:
        """
        try:
            if default is not None:
                default = float(default)
            else:
                default = None
        except:
            default = 0.0
        super().__init__(default, **additional_arguments)

    @returns_none_if_allowed
    def __call__(self, value):
        try:
            result = float(value)
        except Exception:
            result = self._default_value
        return result

    def validate(self, value_to_validate) -> ValidationResult:
        raise NotImplementedError("validate is not implemented within this version.")


class JsonInteger(ADataType):
    """
    Converts a `Any` to a string using the `str` method.

    Examples:

        >>> any_int = JsonInteger(default=0)
        >>> any_int("13")
        13

    Note:
        For more details search for json-schema.org.

    Args:
        default(Optional[int]):
            The default value for this field.

        checkgroups(Optional[str, List[str]]):
            See ADataType

        primekey(Optional[str]):
            See ADataType

        allows_none(bool):
            See ADataType

        multipleOf (int):
            Numbers can be restricted to a multiple of a given number, using the
                        multipleOf keyword. It may be set to any positive number.

        minimum (int):
            Validated for minimum <= x

        exclusiveMinimum (int):
            Validated for minimum < x

        maximum (int):
            Validated for x >= maximum

                exclusiveMaximum (int)
                        Validated for x > maximum

        checkgroups Optional(List[str]):
            Declares a list of group names, this field is assigned to.
            Default is no group assignment.
    """

    KEYWORDS = [
        "multipleOf",
        "minimum",
        "exclusiveMinimum",
        "maximum",
        "exclusiveMaximum",
    ]

    def __init__(self, default: int = 0, **additional_arguments):
        super().__init__(default, **additional_arguments)

    @returns_none_if_allowed
    def __call__(self, incoming_data: Union[str, int, float]) -> int:
        """
        Converts `incoming_data` into an integer. Valid data can be
        numbers represented as string, int or float.

        Args:
            incoming_data:
                Incoming data to be converted into an integer.

        Raises:
            ValueError:
                If `incoming_data` could not be converted.

        Returns:
            int
        """
        try:
            result = int(incoming_data)
            return result
        except ValueError:
            pass
        potential_float = float(incoming_data)
        return int(potential_float)

    def validate(self, value_to_validate) -> ValidationResult:
        raise NotImplementedError(
            "validate is not implemented within this version. A validation of this type"
            "needs to consider the specific KEYWORDS of the integer described by"
            "jsonschema.org"
        )


class SqlInteger(JsonInteger):
    def __init__(self, default=0, **additional_arguments):
        super().__init__(default=default, **additional_arguments)

    @returns_none_if_allowed
    def __call__(self, incoming_data) -> int:
        integer_value = super().__call__(incoming_data)
        if integer_value < SQL_INT_MINIMUM or SQL_INT_MAXIMUM < integer_value:
            raise ValueError(
                "SQL INT only allows values within the range of {} to {}. "
                "Given value was {}.".format(
                    SQL_INT_MINIMUM, SQL_INT_MAXIMUM, integer_value
                )
            )
        return integer_value


class SqlBoolean(ADataType):
    def __init__(self, default=False, allow_none=False, **additional_arguments):
        try:
            default = bool(self._decodestring(default))
        except:
            default = False
        super().__init__(default, **additional_arguments)

    @staticmethod
    def _decodestring(value):
        if isinstance(value, str):
            if value == "":
                return False
            if value == "1":
                return True
            if value.lower() in "truewahr":
                return True
            return False
        return value

    @returns_none_if_allowed
    def __call__(self, incoming_data):
        """
        Returns `false` for any number not being 1 and any string not
         being literal 'true' or 'wahr'; else returns true.

        Args:
            incoming_data:

        Raises:
            ValueError:
                If a 'conversion' didn't succeed.

        Returns:
            bool
        """
        result = bool(self._decodestring(incoming_data))
        return result

    def validate(self, value_to_validate) -> ValidationResult:
        raise NotImplementedError("validate is not implemented within this version.")


class JsonString(ADataType):
    """
    Converts a `Any` to a string using the `str` method.

    Note:
        For more details search for json-schema.org.

    Args:
        columnname (str):
            Key's name of the attribute within the source dictionary.

        default (str):
            Default value for this attribute.

        minLength (int):
            Json-schema field defining the items minimum length.

        maxLength (int):
            Json-schema field defining the items maximum length.

        pattern (str):
            Json-schema field defining a regular exression the value
            needs to match to.

        format (str):
            Json-schema field defining a specific format for this value.
    """

    KEYWORDS = ["minLength", "maxLength", "pattern", "format"]

    def __init__(
        self,
        default="",
        minLength: int = None,
        maxLength: int = None,
        pattern: str = None,
        format: str = None,
        **additional_arguments
    ):
        self._format = format
        self._pattern = pattern
        self._maxLength = maxLength
        self._minLength = minLength
        super().__init__(default, **additional_arguments)

    @returns_none_if_allowed
    def __call__(self, data):
        result = str(data)
        if self._maxLength is not None and self._maxLength < len(result):
            result = result[: self._maxLength]
        if self._minLength is not None and self._minLength > len(result):
            raise ValueError(
                "Given `data` is to short. Necessary length is "
                "{}".format(self._minLength)
            )
        return result

    def to_targettype(self, value):
        warnings.warn("Deprecated method; use class-call instead.", DeprecationWarning)
        return self(value)

    def validate(self, value_to_validate) -> ValidationResult:
        raise NotImplementedError(
            "validate is not implemented within this version. A validation of this type"
            "needs to consider the specific KEYWORDS of the string described by"
            "jsonschema.org"
        )


class SqlChar(JsonString):
    """
    Converts a `Any` to a string using the `str` method.

    Note:
        For more details search for json-schema.org.

    Args:
        columnname (str):
            Key's name of the attribute within the source dictionary.

        default (str):
            Default value for this attribute.

        minLength (int):
            Json-schema field defining the items minimum length.

        maxLength (int):
            Json-schema field defining the items maximum length. For
            SqlChar it is set to 255.

        pattern (str):
            Json-schema field defining a regular exression the value
            needs to match to.

        format (str):
            Json-schema field defining a specific format for this value.
    """

    def __init__(
        self,
        default="",
        minLength: int = None,
        pattern: str = None,
        format: str = None,
        **additional_arguments
    ):
        super().__init__(default, **additional_arguments)
        self._format = format
        self._pattern = pattern
        self._maxLength = SQLCHAR_MAXIMUM_STRING_LENGTH
        self._minLength = minLength



# -*- coding: utf-8 -*-

# from distutils.core import setup
from setuptools import setup, find_packages
from os import path


this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="jsonmetamodel",
    author="David Scheliga",
    author_email="david.scheliga@gmx.de",
    url="https://gitlab.com/david.scheliga/jsonschemamodel",
    project_urls={
        "Source Code Repository": "https://gitlab.com/david.scheliga/jsonschemamodel",
    },
    description="Creates metaclasses for parameter definition and checking with taking"
    " respect to the jsonschema defined by https://json-schema.org/",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="GNU General Public License v3 (GPLv3)",
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3.6",
    ],
    keywords="jsonschema, dict, interfaces",
    package_dir={"jsonmetamodel": "jsonmetamodel"},
    packages=find_packages(),
    install_requires=[
        "pandas",
        "numpy",
        "augmentedtree>=0.2a0",
        "dicthandling>=0.1b5",
    ],
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
)
